--------------
-- Locals
--------------

local L = AceLibrary("AceLocale-2.2"):new("SRA")
local BS = AceLibrary("Babble-Spell-2.2")
local G = AceLibrary("Gratuity-2.0")
local deformat = AceLibrary("Deformat-2.0")
local roster = AceLibrary("Roster-2.1")
local tablet = AceLibrary("Tablet-2.0")
local paintchips = AceLibrary("PaintChips-2.0")
local sharedmedia = AceLibrary("SharedMedia-1.0")
local db = nil -- will be filled with reference to self.db.profile
local enabledicon = "Interface\\AddOns\\SRA\\textures\\enabled.tga"
local disabledicon = "Interface\\AddOns\\SRA\\textures\\disabled.tga" 
local _G = getfenv(0)

local coloredNames = setmetatable({}, {__index =
	function(self, key)
		local obj = roster:GetUnitObjectFromName(key)
		if not obj then return key end
		self[key] = "|cff" .. paintchips:GetHex(obj.class) .. key .. "|r"
		return self[key]
	end
})

local function SetWHP(f, width, height, p1, relative, p2, x, y)
	if not f then return end

	f:SetWidth(width)
	f:SetHeight(height)

	if p1 then
		f:ClearAllPoints()
		f:SetPoint(p1, relative, p2, x, y )
	end
end

local tankStyles = {
	compact = {
		width = 70, height = 34,
		bar = {
			width = 64, height = 16,
			xoff = 2, yoff = -1,
			point = "TOPLEFT", rpoint = "TOPLEFT",
		},
		name = {
			width = 62, height = 14,
			xoff = 4, yoff = -2,
			point = "TOPLEFT", rpoint = "TOPLEFT",
		},
		raidicon = {
			width = 14, height = 14,
			xoff = 5, yoff = 2,
			point = "BOTTOMLEFT", rpoint = "BOTTOMLEFT",
		},
		health = {
			width = 32, height = 14,
			xoff = -4, yoff = 2,
			point = "BOTTOMRIGHT", rpoint = "BOTTOMRIGHT"
		}
	},
	wide = {
		width = 128, height = 21,
		bar = {
			width = 120, height = 16,
			xoff = 4, yoff = 0,
			point = "LEFT", rpoint = "LEFT",
		},
		name = {
			width = 62, height = 14,
			xoff = 22, yoff = 0,
			point = "LEFT", rpoint = "LEFT",
		},
		raidicon = {
			width = 14, height = 14,
			xoff = 5, yoff = 0,
			point = "LEFT", rpoint = "LEFT",
		},
		health = {
			width = 40, height = 14,
			xoff = -4, yoff = 0,
			point = "RIGHT", rpoint = "RIGHT"
		}
	}
}

-----
-- local global variables
-----
local commands = {} -- contains all Comm commands and their handlers
local reincarnationrank = 0 -- Rank of reincarnation talent if applicable
local cooldowns = {} -- CoolDowns status, contains players, spell, and when the cooldown for that spell ends.
local durability = {} -- Durability status table, contains players and their durability + broken items
local maintanktable = {} -- Contains maintanklist
local preftanktable -- set as a raference to self.db.profile.preftanktable for fast lookup, contains preferred tanktable
local smarttanktable = {} -- Contains tanks set by smartTank in a raid or the entire 5 man party
local tanktable = {} -- Contains combined 3 tank tables. This is the actual list in use.
local playerName -- filled with UnitName("player") when we initialize
local readychecks = {} -- Ready Check result table, contains the results of the last/current ready check
local itemchecks = {} -- Item Check result table, contains the results of item checks made this session
local itemtime = {} -- Item Check last response table
local grouped = nil	-- Toggle which indicates if we're in a group or not, controlled with LeftGroup and JoinedGroup functions
local tanks = {} -- Table with the actual currently in use tanks
local mergeTankCombat = nil -- Toggle set if the tanks were updated during combat and should be merged after combat.
local clearTankCombat = nil -- Toggle set if the tanks should be cleared during combat. Will be done after combat.
local secureFrames = {} -- table of our registered secure frames

------
-- Clickcasting global
------

ClickCastFrames = ClickCastFrames or {}

--------------
-- Addon Declaration
--------------

SRA = AceLibrary("AceAddon-2.0"):new("AceEvent-2.0", "AceHook-2.1", "AceDB-2.0", "AceConsole-2.0", "AceComm-2.0", "FuBarPlugin-2.0")
local addon = SRA

local defaults = {
	preftanktable = {},
	durabilitylow = 20,
	promotes = {},
	keyword = nil,
	tanks = {
		type = "SRA",
		style = "compact",
		deficit = nil,
		classcolor = true,
		inversebar = nil,
		enemycolor = "ff0000",
		enemybarcolor = "cc2200",
		coloraggro = true,
		targethighlight = true,
		backdrop = true,
		tooltips = true,
		growup = nil,
		reversed = nil,
		backwards = nil,
		alpha = 1,
		scale = 1,
	},
}

local options = {
	type = "group",
	handler = SRA,
	args = {
		[L["Keyword"]] = {
			type = "text",
			name = L["Keyword"],
			desc = L["Invitation Keyword"],
			set = "SetKeyword",
			get = function() return SRA.db.profile.keyword end,
			usage = L["<keyword> or 'off'"],
		},
		[L["Disband"]] = {
			type = "execute",
			name = L["Disband"],
			desc = L["Disband the group."],
			func = "Disband",
		},
		[L["Promote"]] = {
			type = "text",
			name = L["Promote"],
			desc = L["Set/Unset Auto-Promote."],
			set = "SetAutoPromote",
			get = false,
			usage = L["<name>"],
		},
		[L["Invite"]] = {
			type = "text",
			name = L["Invite"],
			desc = L["Invite guildmembers by level."],
			set = "InviteGuild",
			get = false,
			usage = L["<level>"],
		},
		[L["Item"]] = {
			type = "text",
			name = L["Check Item"],
			desc = L["Check the group for an item."],
			set = "RunItemCheck",
			get = false,
			usage = L["<item>"],
		},
		[L["Tanks"]] = {
			type = "group",
			name = L["Tanks"],
			desc = L["Tank Settings"],
			args = {
				[L["Set MainTank"]] = {
					type = "text",
					name = L["Set MainTank"],
					desc = L["Set a MainTank"],
					set = "SetMainTank",
					get = false,
					usage = L["<name>"],
					disabled = function() return not SRA:IsPromoted() end,
				},
				[L["Remove MainTank"]] = {
					type = "text",
					name = L["Remove MainTank"],
					desc = L["Remove a MainTank"],
					set = "RemoveMainTank",
					get = false,
					usage = L["<name>"],
					disabled = function() return not SRA:IsPromoted() end,					
				},
				[L["Frames"]] = {
					type = "group",
					name = L["Frames"],
					desc = L["Frame Settings"],
					disabled = function() return InCombatLockdown() end,
					args = {
						[L["Alpha"]] = {
							name = L["Alpha"],
							desc = L["Set Alpha"],
							type = "range",
							min = 0, max = 1, step = .05,
							get = function() return SRA.db.profile.tanks.alpha end,
							set = function(v)
									SRA.db.profile.tanks.alpha = v
									SRA.tankFrame:SetAlpha(v)
								end,
						},
						[L["Scale"]] = {
							name = L["Scale"],
							desc = L["Set Scale"],
							type = "range",
							min = 0, max = 2, step = .05,
							get = function() return SRA.db.profile.tanks.scale end,
							set = function(v)
									SRA.db.profile.tanks.scale = v
									SRA.tankFrame:SetScale(v)
								end,						
						},
						[L["Class Color"]] = {
							name = L["Class Color"],
							desc = L["Color healthbars by class."],
							type = "toggle",
							get = function() return SRA.db.profile.tanks.classcolor end,
							set = function(v) SRA.db.profile.tanks.classcolor = v end,
						},
					},
				},
			},
		},
	},
}

--------------
-- Addon Initialization
--------------

function addon:OnInitialize()
	local revision = tonumber(("$Revision: 30841 $"):sub(12, -3)) or 1
	if not self.version then self.version = "1.0" end
	self.version = self.version .. "." .. revision
	self.revision = revision

	_G.SRA_VERSION = self.version
	_G.SRA_REVISION = self.revision

	self:RegisterDB("SRADB", "SRAPerCharDB")
	self:RegisterDefaults("profile", defaults )

	self:RegisterChatCommand( {"/SRA"}, options, "SRA")
	self.OnMenuRequest = options
	
	self.hasNoColor = true
	self.hasIcon = true
end

function addon:OnEnable()
	db = self.db.profile

	preftanktable = db.preftanktable -- this should become a reference right?
	playerName = UnitName("player")
	self:SetIcon(disabledicon)

	-- Setup the frames
	self:InitializeFrames()
	
	-- Register the Comm commands we're interested in
	self:SetCommPrefix("SRA")
	self:RegisterComm( "SRA", "GROUP", "OnComm" )
	self:RegisterCommands()
	
	-- Cooldowns 
	local c = select(2, UnitClass("player"))
	if c == "DRUID" or c == "WARLOCK" or c == "PALADIN" then
		self:RegisterEvent( "UNIT_SPELLCAST_SUCCEEDED" )
	elseif c == "SHAMAN" then
		self:CheckTalents()
		self:SecureHook( "UseSoulstone" )
		self:RegisterEvent( "CHARACTER_POINTS_CHANGED", "CheckTalents" )
	end

	-- Durability
	self:RegisterEvent( "PLAYER_DEAD", "CheckDurability" )
	self:RegisterEvent( "ZONE_CHANGED_NEW_AREA", "CheckDurability" )

	-- Ready Check
	self:SecureHook( "DoReadyCheck" )
	self:SecureHook( "ConfirmReadyCheck" )
	self:RegisterEvent( "READY_CHECK" )

	-- KeyBindings
	self:RegisterEvent( "UPDATE_BINDINGS" )

	-- Keyword invitation
	self:RegisterEvent( "CHAT_MSG_WHISPER" )

	-- Misc
	self:RegisterEvent( "CHAT_MSG_SYSTEM" )
	self:RegisterEvent( "RosterLib_RosterUpdated" )

	-- Leave combat detection for secure updates
	self:RegisterEvent( "PLAYER_REGEN_ENABLED" )
	
	-- make sure we start with a clean slate.
	self:LeftGroup()
end

function addon:Disable()
	self:SetIcon(disabledicon)
end

-------------
-- Comm Handling
-------------

function addon:SendComm(command, ...)
	self:SendCommMessage("GROUP", command, ...)
	self:OnComm(self.commPrefix, playerName, "GROUP", command, ...)
end

function addon:OnComm(prefix, sender, distribution, command, ...)
	if distribution ~= "GROUP" or type(command) ~= "string" then return end
	if type(commands[command]) == "function" then
		commands[command]( self, sender, ... )
	end
end

function addon:RegisterCommand( command, func )
	if not command then return end
	commands[command] = func
end

function addon:RegisterCommands()
	self:RegisterCommand( "CoolDown", self.HandleCooldown )
	self:RegisterCommand( "Durability", self.HandleDurability )
	self:RegisterCommand( "ItemCheck", self.HandleItemCheck )
	self:RegisterCommand( "ItemResponse", self.HandleItemResponse ) 
	self:RegisterCommand( "MainTankSet", self.HandleMainTankSet )
	self:RegisterCommand( "MainTankRemove", self.HandleMainTankRemove )
	self:RegisterCommand( "ReadyCheckResponse", self.HandleReadyCheckResponse )
	self:RegisterCommand( "ReadyCheckEnd", self.HandleReadyCheckEnd )
	self:RegisterCommand( "ReadyCheckUpdate", self.HandleReadyCheckUpdate )
end

function addon:HandleCoolDown( sender, spell, cooldown )
	local t = time() + tonumber(cooldown)
	table.insert(cooldowns, {
		name = sender,
		spell = spell,
		time = t
	})
end

function addon:HandleDurability( sender, perc, broken )
	durability[sender] = {
		perc = perc,
		broken = broken
	}
end

function addon:HandleItemCheck( sender, item )
	if not item then return end
	local numitems = GetItemCount(item)
	if numitems and numitems >= 0 then
		self:SendComm("ItemResponse", item, numitems )
	else
		self:SendComm("ItemResponse", item, 0)
	end
end

function addon:HandleItemResponse( sender, item, numitems )
	if not itemchecks[item] then
		itemchecks[item] = {}
	end
	itemchecks[item][sender] = numitems
	itemtime[item] = time() -- store last responsetime
	self:UpdateItemCheckFrame()
end

function addon:HandleMainTankSet( sender, maintank )
	if not self:IsPromoted( sender ) then return end
	for k, v in ipairs( maintanktable ) do
		if v == maintank then
			table.remove( maintanktable, k)
		end
	end
	table.insert( maintanktable, maintank )
	self:MergeTankTables()
end

function addon:HandleMainTankRemove( sender, maintank )
	if not self:IsPromoted( sender ) then return end
	for k, v in ipairs( maintanktable ) do
		if v == maintank then
			table.remove( maintanktable, k )
		end
	end
	self:MergeTankTables()
end

-- Response sent by a member
function addon:HandleReadyCheckResponse( sender, ready )
	readychecks[sender] = ready
	self:UpdateReadyCheckFrame()
end

-- Update send for a member by the leader
function addon:HandleReadyCheckUpdate( sender, member, ready )
	readychecks[member] = ready
	self:UpdateReadyCheckFrame()
end

-- Leader has received the results.
function addon:HandleReadyCheckEnd( sender, afkers )
	-- Make sure we tag the afkers as afk.
	-- Everyone we received nothing from is ready. Blizzard ready checks can suck my balls.
	if afkers then
		for k, v in pairs({strsplit(",", afkers)}) do
			v = v:trim()
			readychecks[v] = -1
		end
	end
	for unit in roster:IterateRoster() do 
		if not readychecks[unit.name] then
			readychecks[unit.name] = 1
		elseif readychecks[unit.name] == -1 then
			readychecks[unit.name] = nil
		end
	end
	self:UpdateReadyCheckFrame()
	-- Schedule closing of the readycheck thing or something.
	self:ScheduleEvent("SRACloseReady", function() self.readycheckFrame:Hide() end, 3 )
end

------------
-- Event Handling
------------

function addon:PLAYER_REGEN_ENABLED()
	if updateTankCombat then
		updateTankCombat = nil
		self:MergeTankTables()
	end
	if clearTankCombat then
		clearTankCombat = nil
		self:ClearTanks()
	end
end

function addon:CHAT_MSG_WHISPER( msg, author )
	if type(msg) == "string" and db.keyword and msg:lower() == db:lower() then
		if GetNumPartyMembers() == 4 and GetNumRaidMembers() == 0 then ConvertToRaid() end
		if GetNumRaidMembers() == 40 then
			SendChatMessage( L["Sorry, the group is full."], "WHISPER", nil, author)
		else
			self:Print( L["Inviting: %s"]:format( author ) )
			InviteUnit(author)
		end
	end
end

function addon:CHAT_MSG_SYSTEM( msg )
	-- check for ready check system message responses
	-- RAID_MEMBERS_AFK = "The following players are AFK: %s";
	-- RAID_MEMBER_NOT_READY = "%s is not ready";
	-- READY_CHECK_ALL_READY = "Everyone is Ready";
	-- READY_CHECK_MESSAGE = "%s has initiated a ready check.\n|cffffffffAre you ready?|r";
	-- READY_CHECK_NO_AFK = "No players are AFK";
	-- READY_CHECK_START = "Starting ready check...";
	-- READY_CHECK_YOU_WERE_AFK = "You were AFK for a ready check";
	
	local member = deformat(msg, RAID_MEMBERS_AFK)
	if member or msg:find( READY_CHECK_NO_AFK ) or msg:find( READY_CHECK_ALL_READY ) then
		-- ready check has ended
		self:SendComm("ReadyCheckEnd", member)
		return -- we're done here
	else
		member = deformat( msg, RAID_MEMBER_NOT_READY )
		if member and not readychecks[member] then
			self:SendComm("ReadyCheckUpdate", member, 0 )
			return -- we're done here
		end
	end
	if UnitInRaid("player") and IsRaidLeader() then
		member = deformat( msg, ERR_RAID_MEMBER_ADDED_S )
		if member then
			member = member:lower()
			if db.promotes[member] then
				self:Print( L["Autopromoting: %s"]:format( member ) )
				self:ScheduleEvent( PromoteToAssistant, 2, member )
			end
		elseif msg == ERR_NEW_LEADER_YOU or msg == ERR_RAID_YOU_JOINED then
			local counter = 1
			local unit
			for k, v in pairs(db.promotes) do
				unit = roster:GetUnitObjectFromName(k)
				if v and unit and unit.rank == 0 then
					self:Print( L["Autopromoting: %s"]:format( k ) )
					self:ScheduleEvent( PromoteToAssistant, counter, k )
					counter = counter + 1
				end
			end
		end
	end
	
end

function addon:UNIT_SPELLCAST_SUCCEEDED( unit, spell, rank )
	if spell == BS["Rebirth"] then
		self:SendComm("CoolDown", "Rebirth", 30)
	elseif spell == BS["Soulstone Resurrection"] then
		self:SendComm("CoolDown", "Soulstone Resurrection", 30)
	elseif spell == BS["Divine Intervention"] then
		self:SendComm("CoolDown", "Divine Intervention", 60 )
	end
end

function addon:CheckDurability()
	local perc, cur, max, broken, imin, imax = 0, 0, 0, 0, 0, 0
	for i=1,18 do
		G:Erase()
		G:SetInventoryItem("player", i)
		imin, imax = G:FindDeformat(DURABILITY_TEMPLATE)
		if imin and imax then
			imin, imax = tonumber(imin), tonumber(imax)
			if imin == 0 then broken = broken + 1 end
			cur = cur + imin
			max = max + imax
		end
		
	end
	perc = math.floor(cur / max * 100)
	self:SendComm("Durability", perc, broken)
end

function addon:READY_CHECK()
	-- ready check being done
	readychecks = {}
	self:ScheduleEvent("SRACloseReady", function() self.readycheckFrame:Hide() end, 35 )
	local f = self.readycheckFrame
	self:SetOtraviHeight( f )
	f.rightcolumn:SetText( "" )
	f.leftcolumn:SetText( "" )
	if ReadyCheckFrame and ReadyCheckFrame:IsVisible() then
		f:SetPoint( "BOTTOM", ReadyCheckFrame, "TOP" )
	else
		f:SetPoint( "CENTER", UIParent, "CENTER" )
	end
	f:Show()
	self:UpdateReadyCheckFrame()
end

function addon:UPDATE_BINDINGS()
	-- reset the bindings
end

function addon:RosterLib_RosterUpdated()
	if not grouped and ( GetNumRaidMembers() > 0 or GetNumPartyMembers() > 0 ) then
		self:JoinedGroup()
	elseif grouped and GetNumRaidMembers() == 0 and GetNumPartyMembers() == 0 then
		self:LeftGroup()
	end
end

------------
-- Hook Handling
------------

function addon:UseSoulstone()
	local text = HasSoulstone()
	if text and text == BS["Reincarnation"] then
		local cooldown = 60 - (reincarnationrank*10)
		self:SendComm("Reincarnation", cooldown)
	end
end

function addon:DoReadyCheck()
	-- when we perform a readycheck ourselves we don't get the READY_CHECK event, so we trigger the function manually.
	self:READY_CHECK()
	self:SendComm("ReadyCheckResponse", 1 )
end

function addon:ConfirmReadyCheck( ready )
	local resp = ready and ready or 0
	self:SendComm("ReadyCheckResponse", resp )
end

-----------
-- Utility Functions
-----------

function addon:CheckTalents()
	local tab, talent, name, rank

	reincarnationrank = 0

	for tab = 1, GetNumTalentTabs(), 1 do
		for talent = 1, GetNumTalents(tab), 1 do
			name, _, _, _, rank = GetTalentInfo(tab, talent)
			if name == BS["Improved Reincarnation"] then
				reincarnationrank = rank
				break
			end
		end
		if reincarnationrank ~= 0 then
			break
		end 
	end
end

function addon:MergeTankTables()
	-- combat check, if we're in combat we don't do jack shit.
	if InCombatLockdown() then
		mergeTankCombat = true
		return
	end
	-- FIXME: current merge is just a simple tablecopy, because of lack of implementation of other stuff :p
	for k, v in ipairs( tanks ) do
		table.remove( tanks, k )
	end
	-- we're not out of combat so it's safe to merge the tank tables.
	for k, v in ipairs( maintanktable ) do
		table.insert( tanks, v )
	end
	if db.tanks.type == "SRA" then
		if #tanks > 0 then
			self.tankFrame:Show()
			self.tankheaderFrame:SetAttribute( "nameList", table.concat(tanks,",") )
			self.tankheaderFrame:Show()
		else
			self.tankFrame:Hide()
			self.tankheaderFrame:Hide()
		end
	else
		-- optionally push the tanks to agUF and other unitframes here.
	end
end

function addon:ClearTanks()
	if InCombatLockdown() then
		clearTankCombat = true
		return
	end
	for k, v in ipairs( tanks ) do
		table.remove( tanks, k )
	end
	-- clear tanks from other UFs  or hide the raidheader for SRA
	if db.tanks.type == "SRA" then
		self.tankFrame:Hide()
		self.tankheaderFrame:Hide()
	else
		-- clear stuff for other UFs
	end
end

function addon:UpdateTanks()
	for k, f in pairs( secureFrames ) do
		self:UpdateTankFrame( f )
		self:UpdateTankFrame( f.targetframe )
		self:UpdateTankFrame( f.targettargetframe )
	end
end

function addon:UpdateCoolDowns()
	local t = time()
	local s
	for k, v in pairs( cooldowns ) do
		if v.time <= t then
			cooldowns[k] = nil
		else
			s = v.time - t
			cooldowns[k].timeleft = s
		end
	end
end

function addon:LeftGroup()
	if not grouped then return end
	grouped = nil
	self:SetIcon(disabledicon)
	self.updateFrame:SetScript( "OnUpdate", nil )
	self:ClearTanks()
end

function addon:JoinedGroup()
	if grouped then return end
	grouped = true
	self:SetIcon(enabledicon)
	
	self.updateFrame.update = 0
	self.updateFrame:SetScript( "OnUpdate", function() self:OnUpdate() end )
	self:CheckDurability()
	self:MergeTankTables()
end

function addon:OnUpdate()
	this.update = this.update + arg1
	if this.update >= 0.15 then
		this.update = 0
		if db.tanks.type == "SRA" then self:UpdateTanks() end
		self:UpdateCoolDowns()
		self:UpdateDisplay()
	end
end

-----------
-- GUI Functions
-----------

function addon:InitializeFrames()
	if self.updateFrame then return end
	
	local font = GameFontNormal:GetFont()
	
	-- Frame for our OnUpdate handler
	local f = CreateFrame("Frame", "SRAUpdater", UIParent )
	self.updateFrame = f

	-- Drag handle for Tanks
	f = CreateFrame("Frame", "SRATanks", UIParent)
	self.tankFrame = f
	f:Hide()
	f:SetMovable(true)
	f:SetWidth(150) 
	f:SetHeight(13) 
	f:SetPoint("CENTER", UIParent, "CENTER", 0, 0)
	f:SetAlpha( db.tanks.alpha ) 
	f:SetScale( db.tanks.scale ) 
	f:EnableMouse(true)
	f:SetMovable(true)
	f:RegisterForDrag("LeftButton")
	f:SetScript("OnDragStart", function() if not InCombatLockdown() and IsAltKeyDown() then this:StartMoving() end end)
	f:SetScript("OnDragStop", function() if not InCombatLockdown() then this:StopMovingOrSizing() end end)

	f.text = f:CreateFontString(nil,"OVERLAY")
	f.text:SetFont( font, 12 )
	f.text:SetText(L["Tanks"])
	f.text:SetAllPoints(f)

	-- Secure Header for Tanks
	f = CreateFrame("Frame", "SRARaidTankHeader", self.tankFrame, "SecureRaidGroupHeaderTemplate")
	self.raidtankheaderFrame = f
	f:SetFrameStrata("BACKGROUND")
	f:SetWidth(150)
	f:SetHeight(13)
	self:SetSecureHeader(f)
	f:Show()
	
	f = CreateFrame("Frame", "SRAPartyTankHeader", self.tankFrame, "SecurePartyHeaderTemplate")
	self.partytankheaderFrame = f
	f:SetFrameStrata("BACKGROUND")
	f:SetWidth(150)
	f:SetHeight(13)
	self:SetSecureHeader(f)
	f:Show()

	-- Default to something
	-- FIXME: this should be done depending on party/raid somewhere else in the code
	self.tankheaderFrame = self.raidtankheaderFrame
	
	-- Ready check result frame will positioned above the readycheck frame 
	self.readycheckFrame = self:CreateOtraviFrame()
	self.readycheckFrame.header:SetText(L["Ready Check"])

	-- item check resultframe
	self.itemcheckFrame = self:CreateOtraviFrame()
end

function addon:SetSecureHeader( f )
	f:SetAttribute("minheight",13)
	f:SetAttribute("template","SRATankTemplate")

	f:ClearAllPoints()

	if db.tanks.growup then
		f:SetAttribute("point", "TOP")
		f:SetPoint("BOTTOM", self.tankFrame, "TOP", 0, 0 )
	else
		f:SetAttribute("point", "BOTTOM")
		f:SetPoint("TOP", self.tankFrame,"BOTTOM", 0, 0)
	end
	if db.tanks.reversed then
		f:SetAttribute("sortDir", "ASC")
	else
		f:SetAttribute("sortDir", "DESC")
	end
end

function addon:InitializeTankFrame( f )
	-- Tank
	table.insert( secureFrames, f )
	self:SetupTankFrame( f )
	self:SetTankFrameStyle( f )
	
	-- Target of Tank
	f.targetframe = CreateFrame("Button",f:GetName().."Target",f,"SecureActionButtonTemplate")
	local ft = f.targetframe
	self:SetupTankFrame( ft )
	self:SetTankFrameStyle( ft )
    ft:SetAttribute("useparent-unit", true)
    ft:SetAttribute("unitsuffix", "target")
	
	-- Target of Target of Tank
	f.targettargetframe = CreateFrame("Button",f:GetName().."TargetTarget",f,"SecureActionButtonTemplate")
	local ftt = f.targettargetframe
	self:SetupTankFrame( ftt )
	self:SetTankFrameStyle( ftt )
	ftt:SetAttribute("useparent-unit", true)
	ftt:SetAttribute("unitsuffix", "targettarget")
	
	-- Make sure the Target and Target of Target are setpointed correctly:
	self:SetTankFramePoints( f )
	
	-- Enable Clickcasting for our frames
	ClickCastFrames[f] = true
	ClickCastFrames[ft] = true
	ClickCastFrames[ftt] = true
end

function addon:SetTankFramePoints( f )
	local style = tankStyles[db.tanks.style]
	if db.tanks.backwards then
		SetWHP( f.targetframe, style.width, style.height, "RIGHT", f, "LEFT", 0, 0 )
		SetWHP( f.targettargetframe, style.width, style.height, "RIGHT", f.targetframe, "LEFT", 0, 0 )
	else
		SetWHP( f.targetframe, style.width, style.height, "LEFT", f, "LEFT", style.width, 0 )
		SetWHP( f.targettargetframe, style.width, style.height, "LEFT", f.targetframe, "LEFT", style.width, 0 )
	end
end

function addon:SetOtraviHeight( f )
	local num = GetNumRaidMembers()
	if num == 0 then
		num = GetNumPartyMembers()
	end
	num = ( ceil( num / 2 ) * 14 ) + 50
	f:SetHeight( num )
	f.rightcolumn:SetHeight( num )
	f.leftcolumn:SetHeight( num )
end

function addon:CreateOtraviFrame()
	local font = GameFontNormal:GetFont()
	
	local f = CreateFrame("Frame", nil, UIParent)
	f:Hide()
	f:SetWidth( 325 )
	f:SetHeight( 300)
	f:SetBackdrop({
		bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", tile = true, tileSize = 16,
		edgeFile = "Interface\\AddOns\\SRA\\textures\\otravi-semi-full-border", edgeSize = 32,
		insets = {left = 1, right = 1, top = 20, bottom = 1},
	})
	f:SetBackdropColor(24/255, 24/255, 24/255)
	f:EnableMouse(true)
	f:SetMovable(true)
	f:RegisterForDrag("LeftButton")
	f:SetScript("OnDragStart", function() this:StartMoving() end )
	f:SetScript("OnDragStop", function() this:StopMovingOrSizing() end )
	
	-- header
	f.header = f:CreateFontString(nil,"OVERLAY")
	f.header:SetFont(font, 12)
	f.header:SetWidth(300)
	f.header:SetText("")
	f.header:SetShadowOffset(.8, -.8)
	f.header:SetShadowColor(0, 0, 0, 1)
	f.header:ClearAllPoints()
	f.header:SetPoint("TOP", f, "TOP", 0, -14)
	
	-- close button
	f.close = f:CreateTexture(nil, "ARTWORK")
	f.close:SetTexture("Interface\\AddOns\\SRA\\Textures\\otravi-close")
	f.close:SetTexCoord(0, .625, 0, .9333)
	f.close:SetWidth(20)
	f.close:SetHeight(14)
	f.close:SetPoint("TOPRIGHT", f, "TOPRIGHT", -7, -15)
	f.closebutton = CreateFrame("Button", nil, f)
	f.closebutton:SetWidth(20)
	f.closebutton:SetHeight(14)
	f.closebutton:SetPoint("CENTER", f.close, "CENTER")
	f.closebutton:SetScript( "OnClick", function() f:Hide() end )
	-- left column
	f.leftcolumn = f:CreateFontString(nil,"OVERLAY")
	f.leftcolumn:SetFont(font, 12)
	f.leftcolumn:SetWidth(175)
	f.leftcolumn:SetHeight(300)
	f.leftcolumn:SetJustifyV("TOP")
	f.leftcolumn:SetText("")
	f.leftcolumn:ClearAllPoints()
	f.leftcolumn:SetPoint("TOPLEFT", f, "TOPLEFT", 0, -36)
	-- right column
	f.rightcolumn = f:CreateFontString(nil,"OVERLAY")
	f.rightcolumn:SetFont(font, 12)
	f.rightcolumn:SetWidth(175)
	f.rightcolumn:SetHeight(300)
	f.rightcolumn:SetJustifyV("TOP")
	f.rightcolumn:SetText("")
	f.rightcolumn:ClearAllPoints()
	f.rightcolumn:SetPoint("TOPRIGHT", f, "TOPRIGHT", 0, -36)
	
	return f
end

function addon:SetupTankFrame( f )

	local font = GameFontNormal:GetFont()

	-- general scripting and buttons
	f:SetScript("OnEnter", function() self:OnTankEnter() end)
	f:SetScript("OnLeave", function() GameTooltip:Hide() end)
	f:RegisterForClicks("AnyUp")
	f:SetAttribute("*type1", "target")
	
	-- Tank Statusbar
	f.bar = CreateFrame("StatusBar", nil, f)
	f.bar:SetMinMaxValues(0,100)

	-- Tank Statusbar background texture, visible when the bar depleats
	f.bar.texture = f.bar:CreateTexture(nil, "BORDER")
	f.bar.texture:SetVertexColor(1, 0, 0, 0.5)

	-- Tank Statusbar text
	f.bar.health = f.bar:CreateFontString(nil, "OVERLAY")
	f.bar.health:SetFont( font, 10 )
	f.bar.health:SetJustifyH("RIGHT")

	-- Tank Name
	f.name = f.bar:CreateFontString(nil, "OVERLAY")
	f.name:SetFont( font, 10 )
	f.name:SetJustifyH("LEFT")

	-- Raid Icons
	f.raidicon = f.bar:CreateTexture(nil, "OVERLAY")
	f.raidicon:SetTexture("Interface\\TargetingFrame\\UI-RaidTargetingIcons")
	f.raidicon:Hide()

end

function addon:SetTankFrameStyle( f )
	local style = tankStyles[db.tanks.style]
	
	f:SetWidth( style.width )
	f:SetHeight( style.height )
	
	f:SetBackdrop({ bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", tile = true, tileSize = 16})
	if db.tanks.backdrop then
		f:SetBackdropColor(0, 0, 0, .5)
	else
		f:SetBackdropColor(0, 0, 0, 0)
	end

	f.bar:SetStatusBarTexture(self:GetStatusBarTexture())
	f.bar.texture:SetTexture(self:GetStatusBarTexture())
	f.bar.texture:SetVertexColor(.5, .5, .5, .5)

	SetWHP(f.bar, style.bar.width, style.bar.height, style.bar.point, f, style.bar.rpoint, style.bar.xoff, style.bar.yoff)
	SetWHP(f.bar.texture, style.bar.width, style.bar.height, style.bar.point, f, style.bar.rpoint, style.bar.xoff, style.bar.yoff)

	SetWHP(f.raidicon, style.raidicon.width, style.raidicon.height, style.raidicon.point, f, style.raidicon.rpoint, style.raidicon.xoff, style.raidicon.yoff)
	SetWHP(f.name, style.name.width, style.name.height, style.name.point, f, style.name.rpoint, style.name.xoff, style.name.yoff)
	SetWHP(f.bar.health, style.health.width, style.health.height, style.health.point, f, style.health.rpoint, style.health.xoff, style.health.yoff)
end

function addon:GetStatusBarTexture()
	-- FIXME: list the correct textures
	return sharedmedia:Fetch('statusbar', "Blizzard")
end

function addon:UpdateTankFrame( f )

	local unit = SecureButton_GetUnit( f )

	if UnitExists( unit ) then
		f.name:SetText( UnitName( unit ) )
		local suffix = f:GetAttribute("unitsuffix")
		if db.tanks.coloraggro and suffix ~= "target" and suffix ~= "targettarget" then
			if UnitExists( unit .. "target" ) then
				f.name:SetTextColor( 1, 0.5, 0.25, 1 )
				if UnitExists( unit .. "targettarget") then
					if UnitIsUnit(unit, unit .. "targettarget") then
						f.name:SetTextColor(0.5, 1, 0.5, 1)
					else
						f.name:SetTextColor(1, 0, 0, 1)
					end
				end
			else
				f.name:SetTextColor( 1, 1, 1, 1)
			end
		else
			 if UnitIsEnemy(unit, "player") then 
				f.name:SetTextColor(select(2, paintchips:GetRGBPercent( db.tanks.enemycolor ) ))
			else
				f.name:SetTextColor( 1, 1, 1, 1)
			end
		end
		
		self:UpdateTankHealth(f, unit)
		self:UpdateTankRaidIcon(f, unit )
		
		if UnitIsUnit( unit, "target") and db.tanks.targethighlight then
			f:SetBackdropColor(1, .84, 0, 1 )
		elseif db.tanks.backdrop then
			f:SetBackdropColor(0, 0, 0, .5)
		else
			f:SetBackdropColor(0, 0, 0, 0)
		end
		
		f.bar.texture:SetVertexColor(.5, .5, .5, .5)
	else
		f.raidicon:Hide()
		f.bar.texture:SetVertexColor(0, 0, 0, 0)
		f:SetBackdropColor( 0,0,0,0 )
		f.name:SetText( "" )
		f.name:SetTextColor(1, 1, 1, 1)
		f.bar.health:SetText( "" )
		f.bar:SetMinMaxValues(0,1)
		f.bar:SetValue(0)
	end
end

function addon:UpdateTankHealth( f, unit )
	local bar = f.bar
	local cur, max = UnitHealth(unit) or 0, UnitHealthMax(unit) or 0
	local perc = cur / max

	bar:SetMinMaxValues(0, max)

	if db.tanks.inversebar then
		bar:SetValue(max - cur)
	else
		bar:SetValue(cur)
	end

	if db.tanks.classcolor then
		if not UnitIsEnemy(unit, "player") then
			local class = select(2, UnitClass( unit ))
			bar:SetStatusBarColor(select(2, paintchips:GetRGBPercent( class ) ) )
		else
			bar:SetStatusBarColor(select(2, paintchips:GetRGBPercent( db.tanks.enemybarcolor ) ) )
		end
	else
		 bar:SetStatusBarColor(self:GetTankHealthColor(perc))
	end

	if db.tanks.deficit then
		local val = max - cur
		if val > 1000 then
			val = ceil(val/100)/10 .. "k"
		elseif val == 0 then
			val = ""
		end
		bar.health:SetText(val)
	else
		 bar.health:SetText(ceil(perc * 100) .. "%")
	end
	bar:Show()
end

function addon:UpdateTankRaidIcon( f, unit )
	local icon = GetRaidTargetIndex(unit)

	if icon then
		SetRaidTargetIconTexture(f.raidicon, icon)
		f.raidicon:Show()
	else
		f.raidicon:Hide()
	end
end

function addon:GetTankHealthColor( perc )
	local r, g

	if perc > 0.5 then
		r = (1.0 - perc) * 2
		g = 1.0
	else
		r = 1.0
		g = perc * 2
	end

	return r, g, 0
end

function addon:OnTankEnter( )
	-- FIXME: display the remove tank move tank up etc buttons if we're out of combat

	if not db.tanks.tooltips then return end
	local unit = SecureButton_GetUnit( this )
	
	GameTooltip_SetDefaultAnchor(GameTooltip, this)

	if unit and GameTooltip:SetUnit(unit) then
		this.updateTooltip = TOOLTIP_UPDATE_TIME
	else
		this.updateTooltip = nil
	end
end

function addon:OnTooltipUpdate()
	tablet:SetHint( L["tablethint"] )
	if not grouped then return end
	local cat, durabilityfound
	for k, v in pairs ( durability ) do
		if v.perc <= db.durabilitylow or v.broken > 0 then
			if roster:GetUnitObjectFromName(k) then
				if not durabilityfound then 
					cat = tablet:AddCategory(
						"columns", 3,
						"text", L["Name"],
						"justify", "LEFT",
						"text2", L["Durability"],
						"justify2", "LEFT",
						"text3", L["Broken"],
						"child_justify", "LEFT",
						"child_justify2", "LEFT",
						"child_justify3", "RIGHT"
					)
				end
				cat:AddLine( "text", coloredNames[k], "text2", v.perc.."%", "text3", v.broken == 0 and L["None"] or v.broken)
				durabilityfound = true
			end
		end
	end
	if #cooldowns > 0 then
		local h,m,s,t,timetext
		cat = tablet:AddCategory(
			"columns", 3,
			"text", L["Name"],
			"justify", "LEFT",
			"text2", L["Spell"],
			"justify2", "LEFT",
			"text3", L["Time"],
			"child_justify", "LEFT",
			"child_justify2", "LEFT",
			"child_justify3", "RIGHT"
		)
		for k, v in pairs( cooldowns ) do
			if roster:GetUnitIDFromName(v.name) then
				t = v.timeleft
				h = floor(t/3600)
				m = t - (h*3600)
				m = floor(m/60)
				s = t - ((h*3600) + (m*60))
				if h > 0 then
					timetext = ("%d:%02d"):format(h, m)
				else
					timetext = ("%d:%02d"):format(m, floor(s))
				end
				cat:AddLine( "text", coloredNames[v.name], "text2", v.spell, "text3", timetext )
			end
		end
	end
end

function addon:UpdateReadyCheckFrame()
	local text = ""
	local name
	local i = 0
	self.readycheckFrame.leftcolumn:SetText("")
	self.readycheckFrame.rightcolumn:SetText("")
	for unit in roster:IterateRoster() do
		i = i + 1
		if i == 21 then
			self.readycheckFrame.leftcolumn:SetText(text)
			text = "" 
		end
		name = unit.name
		if readychecks[name] == 1 then --ready
			text = text .. "|c0000FF00" .. name .. "|r\n"
		elseif readychecks[name] == 0 then -- not ready
			text = text .. "|c00FF0000" .. name .. "|r\n"
		else -- no reply
			text = text .. "|c00CCCCCC" .. name .. "|r\n"
		end
	end
	if i <= 20 then
		self.readycheckFrame.leftcolumn:SetText(text)
	else
		self.readycheckFrame.rightcolumn:SetText(text)
	end
end

function addon:UpdateItemCheckFrame()
	local f = self.itemcheckFrame
	if not f:IsVisible() or
		not f.item or
		not itemchecks[f.item]
	then return end
	local items = itemchecks[f.item]
	local text = ""
	local name
	local num
	local i = 0
	f.leftcolumn:SetText("")
	f.rightcolumn:SetText("")
	for unit in roster:IterateRoster() do
		i = i + 1
		if i == 21 then
			f.leftcolumn:SetText(text)
			text = "" 
		end
		name = unit.name
		num = items[name]
		if not num then -- no reply
			text = text .. "? |c00CCCCCC" .. name .. "|r\n"
		elseif num == 0 then -- 0 items
			text = text .. num .." |c00FF0000" .. name .. "|r\n"
		else -- more than one item
			text = text .. num .." |c0000FF00" .. name .. "|r\n"
		end
	end
	if i <= 20 then
		f.leftcolumn:SetText(text)
	else
		f.rightcolumn:SetText(text)
	end	
	
end

------------
-- Item Check Command
------------
function addon:RunItemCheck(item)
	if not item or item == "" then return end
	local f = self.itemcheckFrame
	self:SetOtraviHeight( f )
	f.item = item
	f.header:SetText(L["Item Check - %s"]:format( item ))
	f.leftcolumn:SetText("")
	f.rightcolumn:SetText("")
	f:SetPoint( "CENTER", UIParent, "CENTER" )
	f:Show()
	self:UpdateItemCheckFrame()
	self:SendComm( "ItemCheck", item )
end

------------
-- Invitation handling, pretty much a straight copy from oRA2
------------
function addon:InviteGuild(level)
	if not IsInGuild() then
		self:Print(ERR_GUILD_PLAYER_NOT_IN_GUILD)
		return
	end
	if GetNumRaidMembers() == 0 then
		self:Print(ERR_NOT_IN_RAID)
		return
	end

	if not level or not tonumber(level) or level == "" then level = MAX_PLAYER_LEVEL end
	level = tonumber(level)

	SendChatMessage(L["All level %d or higher characters will be invited to raid in 10 seconds. Please leave your groups."]:format(level), "GUILD")
	self:ScheduleEvent(self.DoGuildInvites, 10, self, level)
end

-- actual inviting, expects a number for argument, which is guaranteed by InviteGuild()
function addon:DoGuildInvites(level)
	local offline = GetGuildRosterShowOffline()
	local selection = GetGuildRosterSelection()
	SetGuildRosterShowOffline(0)
	SetGuildRosterSelection(0)
	GetGuildRosterInfo(0)

	local numGuildMembers = GetNumGuildMembers()
	local name, unitLevel, online
	for i = 1, numGuildMembers, 1 do
		name, _, _, unitLevel, _, _, _, _, online = GetGuildRosterInfo(i)
		if level <= unitLevel and name ~= UnitName("player") and online then
			InviteUnit(name)
		end
	end

	SetGuildRosterShowOffline(offline)
	SetGuildRosterSelection(selection)
end

function addon:SetKeyword( keyword ) 
	if keyword == nil or keyword == "" or keyword:lower() == L["off"] then
		db.keyword = nil
		self:Print( L["Keyword inviting disabled."] )
	else
		db.keyword = keyword
		self:Print( L["Invitation keyword set to: %s"]:format( keyword ) )
		self:Print( L["To turn off keyword inviting set it to 'off'."])
	end
end

function addon:Disband()
	if UnitInRaid("player") and ( IsRaidLeader() or IsRaidOfficer() ) then
		local name, online
		SendChatMessage( L["Group disbanding on request by: %s"]:format( playerName ), "RAID")
		for i = 1, GetNumRaidMembers(), 1 do
			name, _, _, _, _, _, _, online = GetRaidRosterInfo(i)
			if online and name ~= UnitName("player") then
				UninviteUnit(name)
			end
		end
		LeaveParty()
	elseif not UnitInRaid("player") and GetNumPartyMembers() > 0 and IsPartyLeader() then
		local name, online
		SendChatMessage( L["Group disbanding on request by: %s"]:format( playerName ), "PARTY")
		for i = GetNumPartyMembers(), 1, -1 do
			if GetPartyMember(i) then
				UninviteUnit( "party"..i )
			end
		end
		LeaveParty()
	end
end

function addon:SetAutoPromote(pname)
	if pname and pname ~= "" then
		pname = pname:lower()
		if db.promotes[pname] then
			db.promotes[pname] = nil
			self:Print( L["Disabling Auto-Promote for: %s"]:format( pname ) )
			local unit = roster:GetUnitObjectFromName(pname)
			if IsRaidLeader() and unit then
				self:Print( L["Demoting: %s"]:format(pname) )
				DemoteAssistant( pname )
			end
		else
			db.promotes[pname] = true
			self:Print( L["Enabling Auto-Promote for: %s"]:format( pname ) )
			local unit = roster:GetUnitObjectFromName(pname)
			if IsRaidLeader() and unit then
				self:Print( L["Autopromoting: %s"]:format( pname ) )
				PromoteToAssistant( pname )
			end
		end
	end
end

function addon:IsPromoted( name )
	if not name then name = playerName end
	if GetNumRaidMembers() > 0 then 
		local unit = roster:GetUnitObjectFromName(name)
		if unit and unit.rank > 0 then
			return true
		end
	elseif GetNumPartyMembers() > 0 then
		local leaderi = GetPartyLeaderIndex()
		if leaderi == 0 then
			if playerName == name then return true end
		elseif leaderi > 0 then
			if playername == UnitName( "party"..leaderi ) then return true end
		end
	end
	return false
end

function addon:SetMainTank( name )
	if not self:IsPromoted() then return end
	if not name then name = UnitName("target") end
	if not name then return end
	local unit = roster:GetUnitObjectFromName( name )
	if not unit then return end
	-- should have verified enough. GO GO GO
	self:SendComm("MainTankSet", name )
end

function addon:RemoveMainTank( name )
	if not self:IsPromoted() then return end
	if not name then name = UnitName("target") end
	if not name then return end
	local unit = roster:GetUnitObjectFromName( name )
	if not unit then return end
	-- should have verified enough. GO GO GO
	self:Sendcomm("MainTankRemove", name )
end
